@@ -1,70 +1,107 @@
-"""
-Base class for module.
-"""
+"""Base class for Module and ConstantsModule."""
 
-from contextlib import suppress
+import ast
 import datetime
-from keyword import iskeyword
-import os
-from pathlib import Path
-import shutil
 import socket
-from tempfile import TemporaryDirectory
+from contextlib import suppress
+from keyword import iskeyword
+from pathlib import Path
 from types import CodeType
-from typing import Dict, List, Optional, Set, Tuple, Union
-
-import importlib_metadata
+from typing import Dict, List, Optional, Sequence, Set, Tuple, Union
 
+from ._compat import importlib_metadata
+from .common import TemporaryPath
 from .exception import ConfigError
 
-
 __all__ = ["ConstantsModule", "Module"]
 
 
 class DistributionCache(importlib_metadata.PathDistribution):
     """Cache the distribution package."""
 
-    _cachedir = TemporaryDirectory(prefix="cxfreeze-")
+    _cachedir = TemporaryPath()
 
     @staticmethod
-    def at(path):
+    def at(path: Union[str, Path]):
         return DistributionCache(Path(path))
 
     at.__doc__ = importlib_metadata.PathDistribution.at.__doc__
 
     @classmethod
-    def from_name(cls, name):
+    def from_name(cls, name: str):
         distribution = super().from_name(name)
+
         # Cache dist-info files in a temporary directory
-        temp_dir = Path(cls._cachedir.name)
-        dist_dir = None
-        files = distribution.files or []
-        prep = importlib_metadata.Prepared(distribution.name)
-        normalized = prep.normalized
-        legacy_normalized = prep.legacy_normalized
-        for file in files:
-            # only existing dist-info files
-            if (
-                not file.match(f"{name}-*.dist-info/*")
-                and not file.match(f"{distribution.name}-*.dist-info/*")
-                and not file.match(f"{normalized}-*.dist-info/*")
-                and not file.match(f"{legacy_normalized}-*.dist-info/*")
-            ):
-                continue
-            src_path = file.locate()
-            if not src_path.exists():
-                continue
-            dst_path = temp_dir / file.as_posix()
-            if dist_dir is None:
-                dist_dir = dst_path.parent
-                dist_dir.mkdir(exist_ok=True)
-            shutil.copy2(src_path, dst_path)
-        if dist_dir is None:
+        normalized_name = getattr(distribution, "_normalized_name", None)
+        if normalized_name is None:
+            normalized_name = importlib_metadata.Prepared.normalize(name)
+        source_path = getattr(distribution, "_path", None)
+        if source_path is None:
+            mask = f"{normalized_name}-{distribution.version}*-info"
+            source_path = list(distribution.locate_file("").glob(mask))[0]
+        if not source_path.exists():
             raise importlib_metadata.PackageNotFoundError(name)
-        return cls.at(dist_dir)
+
+        target_name = f"{normalized_name}-{distribution.version}.dist-info"
+        target_path = cls._cachedir.path / target_name
+        target_path.mkdir(exist_ok=True)
+
+        purelib = None
+        if source_path.name.endswith(".dist-info"):
+            for source in source_path.iterdir():  # type: Path
+                target = target_path / source.name
+                target.write_bytes(source.read_bytes())
+        elif source_path.is_file():
+            # old egg-info file is converted to dist-info
+            target = target_path / "METADATA"
+            target.write_bytes(source_path.read_bytes())
+            purelib = (source_path.parent / (normalized_name + ".py")).exists()
+        else:
+            # Copy minimal data from egg-info directory into dist-info
+            source = source_path / "PKG-INFO"
+            if source.is_file():
+                target = target_path / "METADATA"
+                target.write_bytes(source.read_bytes())
+            source = source_path / "top_level.txt"
+            if source.is_file():
+                target = target_path / "top_level.txt"
+                target.write_bytes(source.read_bytes())
+            purelib = not source_path.joinpath("not-zip-safe").is_file()
+
+        cls._write_wheel_distinfo(target_path, purelib)
+        cls._write_record_distinfo(target_path)
+
+        return cls.at(target_path)
 
     from_name.__doc__ = importlib_metadata.PathDistribution.from_name.__doc__
 
+    @staticmethod
+    def _write_wheel_distinfo(target_path: Path, purelib: bool):
+        """Create a WHEEL file if it doesn't exist"""
+        target = target_path / "WHEEL"
+        if not target.exists():
+            project = Path(__file__).parent.name
+            version = importlib_metadata.version(project)
+            root_is_purelib = "true" if purelib else "false"
+            text = [
+                "Wheel-Version: 1.0",
+                f"Generator: {project} ({version})",
+                f"Root-Is-Purelib: {root_is_purelib}",
+                "Tag: py3-none-any",
+            ]
+            target.write_text("\n".join(text))
+
+    @staticmethod
+    def _write_record_distinfo(target_path: Path):
+        """Recreate a minimal RECORD file"""
+        target_name = target_path.name
+        record = []
+        for file in target_path.iterdir():
+            record.append(f"{target_name}/{file.name},,")
+        record.append(f"{target_name}/RECORD,,")
+        target = target_path / "RECORD"
+        target.write_text("\n".join(record), encoding="utf-8")
+
 
 class Module:
     """
@@ -74,14 +111,14 @@ class Module:
     def __init__(
         self,
         name: str,
-        path: Optional[List[Union[Path, str]]] = None,
+        path: Optional[Sequence[Union[Path, str]]] = None,
         file_name: Optional[Union[Path, str]] = None,
         parent: Optional["Module"] = None,
     ):
         self.name: str = name
-        self.path: Optional[List[Path]] = (
-            [Path(p) for p in path] if path else None
-        )
+        self.path: Optional[List[Path]] = None
+        if path:
+            self.path = [Path(p) for p in path]
         self.file = file_name
         self.parent: Optional["Module"] = parent
         self.code: Optional[CodeType] = None
@@ -92,7 +129,7 @@ class Module:
         self.in_import: bool = True
         self.source_is_string: bool = False
         self.source_is_zip_file: bool = False
-        self._in_file_system: int = 1
+        self.in_file_system = 1
         # cache the dist-info files (metadata)
         self.update_distribution(name)
 
@@ -134,7 +171,8 @@ class Module:
             parts.append(f"file={self.file!r}")
         if self.path is not None:
             parts.append(f"path={self.path!r}")
-        return "<Module {}>".format(", ".join(parts))
+        join_parts = ", ".join(parts)
+        return f"<Module {join_parts}>"
 
     @property
     def in_file_system(self) -> int:
@@ -150,7 +188,7 @@ class Module:
 
     @in_file_system.setter
     def in_file_system(self, value: int) -> None:
-        self._in_file_system = value
+        self._in_file_system: int = value
 
 
 class ConstantsModule:
@@ -179,16 +217,15 @@ class ConstantsModule:
                     value = None
                 else:
                     name, string_value = parts
-                    value = eval(string_value)
+                    value = ast.literal_eval(string_value)
                 if (not name.isidentifier()) or iskeyword(name):
                     raise ConfigError(
                         f"Invalid constant name in ConstantsModule ({name!r})"
                     )
                 self.values[name] = value
-        self._dir: TemporaryDirectory = TemporaryDirectory(prefix="cxfreeze")
-        self.module_path: str = os.path.join(self._dir.name, "constants.py")
+        self.module_path: TemporaryPath = TemporaryPath("constants.py")
 
-    def create(self, modules: List[Module]) -> Tuple[str, str]:
+    def create(self, modules: List[Module]) -> Tuple[Path, str]:
         """
         Create the module which consists of declaration statements for each
         of the values.
@@ -210,12 +247,11 @@ class ConstantsModule:
         self.values["BUILD_TIMESTAMP"] = today.strftime(self.time_format)
         self.values["BUILD_HOST"] = socket.gethostname().split(".")[0]
         self.values["SOURCE_TIMESTAMP"] = stamp.strftime(self.time_format)
-        source_parts = []
+        parts = []
         names = list(self.values.keys())
         names.sort()
         for name in names:
             value = self.values[name]
-            source_parts.append(f"{name} = {value!r}")
-        with open(self.module_path, "w", encoding="UTF-8") as file:
-            file.write("\n".join(source_parts))
-        return self.module_path, self.module_name
+            parts.append(f"{name} = {value!r}")
+        self.module_path.path.write_text("\n".join(parts), encoding="utf-8")
+        return self.module_path.path, self.module_name
